//
// Copyright (c) 2021, NVIDIA CORPORATION. All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of NVIDIA CORPORATION nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#include <optix.h>
#include <optix_function_table_definition.h>
#include <optix_stack_size.h>
#include <optix_stubs.h>

#include <cuda_runtime.h>

#include <sampleConfig.h>

#include <sutil/CUDAOutputBuffer.h>
#include <sutil/Exception.h>
#include <sutil/sutil.h>

#include "optixMotionBlurTest.h"

#include <array>
#include <iomanip>
#include <iostream>
#include <string>

#include <sutil/Camera.h>
#include <sutil/Trackball.h>

template <typename T>
struct SbtRecord
{
    __align__(OPTIX_SBT_RECORD_ALIGNMENT) char header[OPTIX_SBT_RECORD_HEADER_SIZE];
    T data;
};

typedef SbtRecord<RayGenData>     RayGenSbtRecord;
typedef SbtRecord<MissData>       MissSbtRecord;
typedef SbtRecord<HitGroupData>   HitGroupSbtRecord;


void configureCamera(sutil::Camera& cam, const uint32_t width, const uint32_t height)
{
    cam.setEye({ 0.0f, 0.0f, 2.0f });
    cam.setLookat({ 0.0f, 0.0f, 0.0f });
    cam.setUp({ 0.0f, 1.0f, 3.0f });
    cam.setFovY(45.0f);
    cam.setAspectRatio((float)width / (float)height);
}


void printUsageAndExit(const char* argv0)
{
    std::cerr << "Usage  : " << argv0 << " [options]\n";
    std::cerr << "Options: --file | -f <filename>      Specify file for image output\n";
    std::cerr << "         --help | -h                 Print this usage message\n";
    std::cerr << "         --dim=<width>x<height>      Set image dimensions; defaults to 512x384\n";
    exit(1);
}


static void context_log_cb(unsigned int level, const char* tag, const char* message, void* /*cbdata */)
{
    std::cerr << "[" << std::setw(2) << level << "][" << std::setw(12) << tag << "]: "
        << message << "\n";
}


static void BuildAS(OptixDeviceContext& context, OptixAccelBuildOptions& accel_options, OptixBuildInput& buildInput, CUdeviceptr& d_output_buffer, OptixTraversableHandle& handle)
{
    OptixAccelBufferSizes gas_buffer_sizes;
    OPTIX_CHECK(optixAccelComputeMemoryUsage(
        context,
        &accel_options,
        &buildInput,
        1, // Number of build inputs
        &gas_buffer_sizes
    ));
    CUdeviceptr d_temp_buffer;
    CUDA_CHECK(cudaMalloc(
        reinterpret_cast<void**>(&d_temp_buffer),
        gas_buffer_sizes.tempSizeInBytes
    ));
    CUDA_CHECK(cudaMalloc(
        reinterpret_cast<void**>(&d_output_buffer),
        gas_buffer_sizes.outputSizeInBytes
    ));

    OPTIX_CHECK(optixAccelBuild(
        context,
        0,                  // CUDA stream
        &accel_options,
        &buildInput,
        1,                  // num build inputs
        d_temp_buffer,
        gas_buffer_sizes.tempSizeInBytes,
        d_output_buffer,
        gas_buffer_sizes.outputSizeInBytes,
        &handle,
        nullptr,            // emitted property list
        0                   // num emitted properties
    ));

    // We can now free the scratch space buffer used during build
    CUDA_CHECK(cudaFree(reinterpret_cast<void*>(d_temp_buffer)));
}

int main(int argc, char* argv[])
{
    std::string outfile;
    int         width = 1024;
    int         height = 768;

    const size_t numInstances = 50;
    const float identityTransform[12] = {
        1,0,0,0,
        0,1,0,0,
        0,0,1,0
    };

    const OptixSRTData identitySRT = {
        1,0,0,0,
        1,0,0,1,
        0,0,0,0,
        1,0,0,0
    };

    for (int i = 1; i < argc; ++i)
    {
        const std::string arg(argv[i]);
        if (arg == "--help" || arg == "-h")
        {
            printUsageAndExit(argv[0]);
        }
        else if (arg == "--file" || arg == "-f")
        {
            if (i < argc - 1)
            {
                outfile = argv[++i];
            }
            else
            {
                printUsageAndExit(argv[0]);
            }
        }
        else if (arg.substr(0, 6) == "--dim=")
        {
            const std::string dims_arg = arg.substr(6);
            sutil::parseDimensions(dims_arg.c_str(), width, height);
        }
        else
        {
            std::cerr << "Unknown option '" << arg << "'\n";
            printUsageAndExit(argv[0]);
        }
    }

    try
    {
        //
        // Initialize CUDA and create OptiX context
        //
        OptixDeviceContext context = nullptr;
        {
            // Initialize CUDA
            CUDA_CHECK(cudaFree(0));

            // Initialize the OptiX API, loading all API entry points
            OPTIX_CHECK(optixInit());

            // Specify context options
            OptixDeviceContextOptions options = {};
            options.logCallbackFunction = &context_log_cb;
            options.logCallbackLevel = 4;
            options.validationMode = OptixDeviceContextValidationMode::OPTIX_DEVICE_CONTEXT_VALIDATION_MODE_ALL;
            // Associate a CUDA context (and therefore a specific GPU) with this
            // device context
            CUcontext cuCtx = 0;  // zero means take the current context
            OPTIX_CHECK(optixDeviceContextCreate(cuCtx, &options, &context));
        }


        //
        // accel handling
        //
        OptixTraversableHandle gas_handle;
        CUdeviceptr            d_gas_output_buffer;
        {
            // Use default options for simplicity.  In a real use case we would want to
            // enable compaction, etc
            OptixAccelBuildOptions accel_options = {};
            accel_options.buildFlags = OPTIX_BUILD_FLAG_NONE;
            accel_options.operation = OPTIX_BUILD_OPERATION_BUILD;

            // Triangle build input: simple list of three vertices
            const std::array<float3, 3> vertices =
            { {
                  { -0.5f, -0.5f, 0.0f },
                  {  0.5f, -0.5f, 0.0f },
                  {  0.0f,  0.5f, 0.0f }
            } };

            const size_t vertices_size = sizeof(float3) * vertices.size();
            CUdeviceptr d_vertices = 0;
            CUDA_CHECK(cudaMalloc(reinterpret_cast<void**>(&d_vertices), vertices_size));
            CUDA_CHECK(cudaMemcpy(
                reinterpret_cast<void*>(d_vertices),
                vertices.data(),
                vertices_size,
                cudaMemcpyHostToDevice
            ));

            // Our build input is a simple list of non-indexed triangle vertices
            const uint32_t triangle_input_flags[1] = { OPTIX_GEOMETRY_FLAG_NONE };
            OptixBuildInput triangle_input = {};
            triangle_input.type = OPTIX_BUILD_INPUT_TYPE_TRIANGLES;
            triangle_input.triangleArray.vertexFormat = OPTIX_VERTEX_FORMAT_FLOAT3;
            triangle_input.triangleArray.numVertices = static_cast<uint32_t>(vertices.size());
            triangle_input.triangleArray.vertexBuffers = &d_vertices;
            triangle_input.triangleArray.flags = triangle_input_flags;
            triangle_input.triangleArray.numSbtRecords = 1;

            BuildAS(context, accel_options, triangle_input, d_gas_output_buffer, gas_handle);
            CUDA_CHECK(cudaFree(reinterpret_cast<void*>(d_vertices)));
        }


        std::vector<OptixTraversableHandle> ias_handles(numInstances);
        std::vector<CUdeviceptr>            d_ias_output_buffers(numInstances);

        std::vector<CUdeviceptr>            d_srt_transforms(numInstances);

        OptixTraversableHandle top_level_ias_handle;
        CUdeviceptr            d_top_level_ias_output_buffer;
        {
            std::vector<OptixInstance> instancesHost;
            instancesHost.reserve(numInstances);

            for (size_t idx = 0; idx < numInstances; idx++)
            {
                OptixInstance i = {};
                i.instanceId = 0;
                i.visibilityMask = 255;
                i.sbtOffset = 0;
                i.flags = OPTIX_INSTANCE_FLAG_NONE;
                i.traversableHandle = gas_handle;
                memcpy(i.transform, identityTransform, 12 * sizeof(float));

                CUdeviceptr d_instance = 0;
                CUDA_CHECK(cudaMalloc(
                    reinterpret_cast<void**>(&d_instance),
                    sizeof(OptixInstance)
                ));
                CUDA_CHECK(cudaMemcpy(
                    reinterpret_cast<void*>(d_instance),
                    &i,
                    sizeof(OptixInstance),
                    cudaMemcpyHostToDevice
                ));

                OptixBuildInput buildInput = {};
                buildInput.type = OPTIX_BUILD_INPUT_TYPE_INSTANCES;
                buildInput.instanceArray.instances = d_instance;
                buildInput.instanceArray.numInstances = 1;
                buildInput.instanceArray.instanceStride = 0;

                OptixAccelBuildOptions accelOptions = {};
                accelOptions.operation = OPTIX_BUILD_OPERATION_BUILD;
                accelOptions.buildFlags = OPTIX_BUILD_FLAG_PREFER_FAST_TRACE;

                BuildAS(context, accelOptions, buildInput, d_ias_output_buffers[idx], ias_handles[idx]);

                CUDA_CHECK(cudaFree(reinterpret_cast<void*>(d_instance)));

                OptixSRTMotionTransform motionTransform = { 0 };
                motionTransform.child = ias_handles[idx];
                motionTransform.motionOptions.flags = OPTIX_MOTION_FLAG_NONE;
                motionTransform.motionOptions.numKeys = 2;
                motionTransform.motionOptions.timeBegin = 0.0f;
                motionTransform.motionOptions.timeEnd = 1.0f;
                memcpy(&motionTransform.srtData[0], &identitySRT, sizeof(OptixSRTData));
                memcpy(&motionTransform.srtData[1], &identitySRT, sizeof(OptixSRTData));

                CUDA_CHECK(cudaMalloc(
                    reinterpret_cast<void**>(&d_srt_transforms[idx]),
                    sizeof(OptixSRTMotionTransform)
                ));

                OPTIX_CHECK(optixConvertPointerToTraversableHandle(
                    context,
                    d_srt_transforms[idx],
                    OPTIX_TRAVERSABLE_TYPE_SRT_MOTION_TRANSFORM,
                    &ias_handles[idx]
                ));

                CUDA_CHECK(cudaMemcpy(reinterpret_cast<void*>(d_srt_transforms[idx]), &motionTransform, sizeof(OptixSRTMotionTransform), cudaMemcpyHostToDevice));

                i.traversableHandle = ias_handles[idx];

                instancesHost.push_back(i);
            }

            CUdeviceptr d_instances = 0;
            CUDA_CHECK(cudaMalloc(
                reinterpret_cast<void**>(&d_instances),
                numInstances * sizeof(OptixInstance)
            ));
            CUDA_CHECK(cudaMemcpy(
                reinterpret_cast<void*>(d_instances),
                instancesHost.data(),
                numInstances * sizeof(OptixInstance),
                cudaMemcpyHostToDevice
            ));

            OptixBuildInput buildInput = {};
            buildInput.type = OPTIX_BUILD_INPUT_TYPE_INSTANCES;
            buildInput.instanceArray.instances = d_instances;
            buildInput.instanceArray.numInstances = numInstances;
            buildInput.instanceArray.instanceStride = 0;

            OptixAccelBuildOptions accelOptions = {};
            accelOptions.operation = OPTIX_BUILD_OPERATION_BUILD;
            accelOptions.buildFlags = OPTIX_BUILD_FLAG_PREFER_FAST_BUILD;

            accelOptions.motionOptions.flags = OPTIX_MOTION_FLAG_NONE;
            accelOptions.motionOptions.numKeys = 2;
            accelOptions.motionOptions.timeBegin = 0.0f;
            accelOptions.motionOptions.timeEnd = 1.0f;

            BuildAS(context, accelOptions, buildInput, d_top_level_ias_output_buffer, top_level_ias_handle);

            CUDA_CHECK(cudaFree(reinterpret_cast<void*>(d_instances)));
        }


        //
        // Create module
        //
        OptixModule module = nullptr;
        OptixPipelineCompileOptions pipeline_compile_options = {};
        {
            OptixModuleCompileOptions module_compile_options = {};
            module_compile_options.optLevel = OPTIX_COMPILE_OPTIMIZATION_LEVEL_0;
            module_compile_options.debugLevel = OPTIX_COMPILE_DEBUG_LEVEL_FULL;

            pipeline_compile_options.usesMotionBlur = true;
            pipeline_compile_options.traversableGraphFlags = OPTIX_TRAVERSABLE_GRAPH_FLAG_ALLOW_ANY;
            pipeline_compile_options.numPayloadValues = 3;
            pipeline_compile_options.numAttributeValues = 3;
            //#ifdef DEBUG // Enables debug exceptions during optix launches. This may incur significant performance cost and should only be done during development.
            pipeline_compile_options.exceptionFlags = OPTIX_EXCEPTION_FLAG_DEBUG | OPTIX_EXCEPTION_FLAG_TRACE_DEPTH | OPTIX_EXCEPTION_FLAG_STACK_OVERFLOW |
                OPTIX_EXCEPTION_FLAG_USER;
            //#else
            //            pipeline_compile_options.exceptionFlags = OPTIX_EXCEPTION_FLAG_NONE;
            //#endif
            pipeline_compile_options.pipelineLaunchParamsVariableName = "params";
            pipeline_compile_options.usesPrimitiveTypeFlags = OPTIX_PRIMITIVE_TYPE_FLAGS_TRIANGLE;

            size_t      inputSize = 0;
            const char* input = sutil::getInputData(OPTIX_SAMPLE_NAME, OPTIX_SAMPLE_DIR, "optixMotionBlurTest.cu", inputSize);

            OPTIX_CHECK_LOG(optixModuleCreate(
                context,
                &module_compile_options,
                &pipeline_compile_options,
                input,
                inputSize,
                LOG, &LOG_SIZE,
                &module
            ));
        }

        //
        // Create program groups
        //
        OptixProgramGroup raygen_prog_group = nullptr;
        OptixProgramGroup miss_prog_group = nullptr;
        OptixProgramGroup hitgroup_prog_group = nullptr;
        {
            OptixProgramGroupOptions program_group_options = {}; // Initialize to zeros

            OptixProgramGroupDesc raygen_prog_group_desc = {}; //
            raygen_prog_group_desc.kind = OPTIX_PROGRAM_GROUP_KIND_RAYGEN;
            raygen_prog_group_desc.raygen.module = module;
            raygen_prog_group_desc.raygen.entryFunctionName = "__raygen__rg";
            OPTIX_CHECK_LOG(optixProgramGroupCreate(
                context,
                &raygen_prog_group_desc,
                1,   // num program groups
                &program_group_options,
                LOG, &LOG_SIZE,
                &raygen_prog_group
            ));

            OptixProgramGroupDesc miss_prog_group_desc = {};
            miss_prog_group_desc.kind = OPTIX_PROGRAM_GROUP_KIND_MISS;
            miss_prog_group_desc.miss.module = module;
            miss_prog_group_desc.miss.entryFunctionName = "__miss__ms";
            OPTIX_CHECK_LOG(optixProgramGroupCreate(
                context,
                &miss_prog_group_desc,
                1,   // num program groups
                &program_group_options,
                LOG, &LOG_SIZE,
                &miss_prog_group
            ));

            OptixProgramGroupDesc hitgroup_prog_group_desc = {};
            hitgroup_prog_group_desc.kind = OPTIX_PROGRAM_GROUP_KIND_HITGROUP;
            hitgroup_prog_group_desc.hitgroup.moduleCH = module;
            hitgroup_prog_group_desc.hitgroup.entryFunctionNameCH = "__closesthit__ch";
            OPTIX_CHECK_LOG(optixProgramGroupCreate(
                context,
                &hitgroup_prog_group_desc,
                1,   // num program groups
                &program_group_options,
                LOG, &LOG_SIZE,
                &hitgroup_prog_group
            ));
        }

        //
        // Link pipeline
        //
        OptixPipeline pipeline = nullptr;
        {
            const uint32_t    max_trace_depth = 4;
            OptixProgramGroup program_groups[] = { raygen_prog_group, miss_prog_group, hitgroup_prog_group };

            OptixPipelineLinkOptions pipeline_link_options = {};
            pipeline_link_options.maxTraceDepth = max_trace_depth;
            OPTIX_CHECK_LOG(optixPipelineCreate(
                context,
                &pipeline_compile_options,
                &pipeline_link_options,
                program_groups,
                sizeof(program_groups) / sizeof(program_groups[0]),
                LOG, &LOG_SIZE,
                &pipeline
            ));

            OptixStackSizes stack_sizes = {};
            for (auto& prog_group : program_groups)
            {
                OPTIX_CHECK(optixUtilAccumulateStackSizes(prog_group, &stack_sizes, pipeline));
            }

            uint32_t direct_callable_stack_size_from_traversal;
            uint32_t direct_callable_stack_size_from_state;
            uint32_t continuation_stack_size;
            OPTIX_CHECK(optixUtilComputeStackSizes(&stack_sizes, max_trace_depth,
                0,  // maxCCDepth
                0,  // maxDCDEpth
                &direct_callable_stack_size_from_traversal,
                &direct_callable_stack_size_from_state, &continuation_stack_size));
            OPTIX_CHECK(optixPipelineSetStackSize(pipeline, direct_callable_stack_size_from_traversal,
                direct_callable_stack_size_from_state, continuation_stack_size,
                4  // maxTraversableDepth
            ));
        }

        //
        // Set up shader binding table
        //
        OptixShaderBindingTable sbt = {};
        {
            CUdeviceptr  raygen_record;
            const size_t raygen_record_size = sizeof(RayGenSbtRecord);
            CUDA_CHECK(cudaMalloc(reinterpret_cast<void**>(&raygen_record), raygen_record_size));
            RayGenSbtRecord rg_sbt;
            OPTIX_CHECK(optixSbtRecordPackHeader(raygen_prog_group, &rg_sbt));
            CUDA_CHECK(cudaMemcpy(
                reinterpret_cast<void*>(raygen_record),
                &rg_sbt,
                raygen_record_size,
                cudaMemcpyHostToDevice
            ));

            CUdeviceptr miss_record;
            size_t      miss_record_size = sizeof(MissSbtRecord);
            CUDA_CHECK(cudaMalloc(reinterpret_cast<void**>(&miss_record), miss_record_size));
            MissSbtRecord ms_sbt;
            ms_sbt.data = { 0.3f, 0.1f, 0.2f };
            OPTIX_CHECK(optixSbtRecordPackHeader(miss_prog_group, &ms_sbt));
            CUDA_CHECK(cudaMemcpy(
                reinterpret_cast<void*>(miss_record),
                &ms_sbt,
                miss_record_size,
                cudaMemcpyHostToDevice
            ));

            CUdeviceptr hitgroup_record;
            size_t      hitgroup_record_size = sizeof(HitGroupSbtRecord);
            CUDA_CHECK(cudaMalloc(reinterpret_cast<void**>(&hitgroup_record), hitgroup_record_size));
            HitGroupSbtRecord hg_sbt;
            OPTIX_CHECK(optixSbtRecordPackHeader(hitgroup_prog_group, &hg_sbt));
            CUDA_CHECK(cudaMemcpy(
                reinterpret_cast<void*>(hitgroup_record),
                &hg_sbt,
                hitgroup_record_size,
                cudaMemcpyHostToDevice
            ));

            sbt.raygenRecord = raygen_record;
            sbt.missRecordBase = miss_record;
            sbt.missRecordStrideInBytes = sizeof(MissSbtRecord);
            sbt.missRecordCount = 1;
            sbt.hitgroupRecordBase = hitgroup_record;
            sbt.hitgroupRecordStrideInBytes = sizeof(HitGroupSbtRecord);
            sbt.hitgroupRecordCount = 1;
        }

        sutil::CUDAOutputBuffer<uchar4> output_buffer(sutil::CUDAOutputBufferType::CUDA_DEVICE, width, height);

        //
        // launch
        //
        {
            CUstream stream;
            CUDA_CHECK(cudaStreamCreate(&stream));

            sutil::Camera cam;
            configureCamera(cam, width, height);

            Params params;
            params.image = output_buffer.map();
            params.image_width = width;
            params.image_height = height;
            params.handle = top_level_ias_handle;
            params.cam_eye = cam.eye();
            cam.UVWFrame(params.cam_u, params.cam_v, params.cam_w);

            CUdeviceptr d_param;
            CUDA_CHECK(cudaMalloc(reinterpret_cast<void**>(&d_param), sizeof(Params)));
            CUDA_CHECK(cudaMemcpy(
                reinterpret_cast<void*>(d_param),
                &params, sizeof(params),
                cudaMemcpyHostToDevice
            ));

            OPTIX_CHECK(optixLaunch(pipeline, stream, d_param, sizeof(Params), &sbt, width, height, /*depth=*/1));
            CUDA_SYNC_CHECK();

            output_buffer.unmap();
            CUDA_CHECK(cudaFree(reinterpret_cast<void*>(d_param)));
        }

        //
        // Display results
        //
        {
            sutil::ImageBuffer buffer;
            buffer.data = output_buffer.getHostPointer();
            buffer.width = width;
            buffer.height = height;
            buffer.pixel_format = sutil::BufferImageFormat::UNSIGNED_BYTE4;
            if (outfile.empty())
                sutil::displayBufferWindow(argv[0], buffer);
            else
                sutil::saveImage(outfile.c_str(), buffer, false);
        }

        //
        // Cleanup
        //
        {
            CUDA_CHECK(cudaFree(reinterpret_cast<void*>(sbt.raygenRecord)));
            CUDA_CHECK(cudaFree(reinterpret_cast<void*>(sbt.missRecordBase)));
            CUDA_CHECK(cudaFree(reinterpret_cast<void*>(sbt.hitgroupRecordBase)));
            CUDA_CHECK(cudaFree(reinterpret_cast<void*>(d_gas_output_buffer)));
            CUDA_CHECK(cudaFree(reinterpret_cast<void*>(d_top_level_ias_output_buffer)));
            for (size_t i = 0; i < numInstances; i++)
            {
                CUDA_CHECK(cudaFree(reinterpret_cast<void*>(d_ias_output_buffers[i])));
                CUDA_CHECK(cudaFree(reinterpret_cast<void*>(d_srt_transforms[i])));
            }

            OPTIX_CHECK(optixPipelineDestroy(pipeline));
            OPTIX_CHECK(optixProgramGroupDestroy(hitgroup_prog_group));
            OPTIX_CHECK(optixProgramGroupDestroy(miss_prog_group));
            OPTIX_CHECK(optixProgramGroupDestroy(raygen_prog_group));
            OPTIX_CHECK(optixModuleDestroy(module));

            OPTIX_CHECK(optixDeviceContextDestroy(context));
        }
    }
    catch (std::exception& e)
    {
        std::cerr << "Caught exception: " << e.what() << "\n";
        return 1;
    }
    return 0;
}
